#include <iostream>
#include <vector>


#include <GL/glew.h>
#include <GL/glut.h>
#include <GL/glx.h>

#include "CL/cl_gl.h"

#define VIENNACL_WITH_OPENCL

#include "viennacl/vector.hpp"
#include "viennacl/matrix.hpp"


#define GL_ERROR_CHECK { GLenum error = glGetError();  if (error != GL_NO_ERROR) std::cout << "GL Error \"" << gluErrorString( error ) << "\" at " << __LINE__ << std::endl; }
#define CL_ERROR_CHECK(error) { if (error != CL_SUCCESS) std::cout << "CL error at " << __LINE__ << std::endl; }


void processNormalKeys(unsigned char key, int x, int y) {

    if (key == 27)
        exit(0);
}


void changeSize(int w, int h) {
    glViewport(0, 0, w, h);
}



unsigned int gl_buffer;


unsigned int gl_texture_0;
unsigned int gl_texture_1;


void renderScene(void) {

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, gl_texture_0);
    
    glBegin(GL_QUADS);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-0.8,-0.8,0.0);
        
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-0.8,0.8,0.0);
        
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(-0.2,0.8,0.0);
        
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(-0.2,-0.8,0.0);
    glEnd();

    glDisable(GL_TEXTURE_2D);
    
    
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, gl_texture_1);
    
    glBegin(GL_QUADS);
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(0.8,0.8,0.0);
        
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(0.8,-0.8,0.0);
        
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(0.2,-0.8,0.0);
        
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(0.2,0.8,0.0);
    glEnd();
    
    glDisable(GL_TEXTURE_2D);
    
    glutSwapBuffers();
}



int main(int argc, char **argv)
{
    ////////////////////////////////////////////////////////////////
    //     OpenGL Setup
    ////////////////////////////////////////////////////////////////
    
    glutInit(&argc, argv);
    
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutInitWindowPosition(100,100);
    glutInitWindowSize(800,400);
    glutCreateWindow("CL - GL");
    
    glewInit();
    

    ////////////////////////////////////////////////////////////////
    //     OpenCL Context Setup
    ////////////////////////////////////////////////////////////////

    
    
    cl_context_properties properties[] = {
        CL_GL_CONTEXT_KHR, (cl_context_properties) glXGetCurrentContext(),
        CL_GLX_DISPLAY_KHR, (cl_context_properties) glXGetCurrentDisplay(),
        CL_CONTEXT_PLATFORM, (cl_context_properties) viennacl::ocl::get_platforms()[0].id(),
        0};
    
    
    cl_device_id my_device = viennacl::ocl::current_device().id();
        
    
    cl_int err;
    
    cl_context my_context = clCreateContext(properties, 1, &my_device, NULL, NULL, &err);
        CL_ERROR_CHECK(err)
        
    cl_command_queue my_queue = clCreateCommandQueue( my_context, my_device, 0, &err );
        CL_ERROR_CHECK(err)
        
        
    ////////////////////////////////////////////////////////////////
    //     ViennaCL Setup
    ////////////////////////////////////////////////////////////////

    viennacl::ocl::setup_context( 1, my_context, my_device, my_queue );
    viennacl::ocl::switch_context( 1 );
        
    
    
    ////////////////////////////////////////////////////////////////
    //     OpenGL Texture Setup
    ////////////////////////////////////////////////////////////////
    
    glGenTextures(1, &gl_texture_0);
    
    
    unsigned int texture_width = 256;
    unsigned int texture_height = 256;
    std::vector<float> texture_buffer(texture_width*texture_height*3);
    
    std::fill( texture_buffer.begin(), texture_buffer.end(), 0 );
    for (unsigned int x = 0; x < texture_width; ++x)
        for (unsigned int y = 0; y < texture_height; ++y)
        {
            unsigned int pos = (y*texture_width + x)*3;
            texture_buffer[pos+0] = float(x)/texture_width;
            texture_buffer[pos+1] = float(y)/texture_height;
        }
        
    glBindTexture(GL_TEXTURE_2D, gl_texture_0);
        GL_ERROR_CHECK
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB32F, texture_width, texture_height, 0, GL_RGB, GL_FLOAT, &texture_buffer[0] );
        GL_ERROR_CHECK
    
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        GL_ERROR_CHECK
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        GL_ERROR_CHECK
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT );
        GL_ERROR_CHECK
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT );
        GL_ERROR_CHECK
    
    
    
    ////////////////////////////////////////////////////////////////
    //     OpenGL Buffer & OpenCL Memory Setup
    ////////////////////////////////////////////////////////////////
    

    glGenBuffers(1, &gl_buffer);
        GL_ERROR_CHECK
    glBindBuffer( GL_PIXEL_UNPACK_BUFFER, gl_buffer );
        GL_ERROR_CHECK
    glBufferData( GL_PIXEL_UNPACK_BUFFER, texture_width*texture_height*3 * sizeof(float), NULL, GL_DYNAMIC_COPY );
        GL_ERROR_CHECK
        
        
    // Link OpenGL buffer to OpenCL memory
    cl_mem cl_buffer = clCreateFromGLBuffer(my_context, CL_MEM_READ_WRITE, gl_buffer, &err);
        CL_ERROR_CHECK(err)
        
        
    // Fill temporay vector with initial data
    std::vector<float> tmp_vec(texture_width*texture_height*3);
    std::fill( tmp_vec.begin(), tmp_vec.end(), 0 );
    for (unsigned int x = 0; x < texture_width; ++x)
        for (unsigned int y = 0; y < texture_height; ++y)
        {
            unsigned int pos = (y*texture_width + x)*3;
            tmp_vec[pos+2] = float(x)/texture_width;
            tmp_vec[pos+1] = float(y)/texture_height;
        }

    
    // Create viennacl vector from OpenCL memory
    viennacl::vector<float> my_vec(cl_buffer, texture_width*texture_height*3);
    
    // Acquire memory object for write read/write operation
    clEnqueueAcquireGLObjects(my_queue, 1, &cl_buffer, 0, NULL, NULL);
    
    // copy CPU data to viennacl
    viennacl::copy( tmp_vec, my_vec );
    // doing some stuff
    my_vec *= 0.5f;
    
    // Release memory object
    clEnqueueReleaseGLObjects(my_queue, 1, &cl_buffer, 0, NULL, NULL);
        
    
    // Creating Texture from OpenGL buffer    
    glGenTextures(1, &gl_texture_1);
        GL_ERROR_CHECK
    glBindTexture( GL_TEXTURE_2D, gl_texture_1 );
        GL_ERROR_CHECK
    
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB32F, texture_width, texture_height, 0, GL_RGB, GL_FLOAT, 0 );
        GL_ERROR_CHECK
    
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        GL_ERROR_CHECK
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        GL_ERROR_CHECK
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT );
        GL_ERROR_CHECK
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT );
        GL_ERROR_CHECK
    
    glBindBuffer( GL_PIXEL_UNPACK_BUFFER, 0 );
    
    

    
    
    // register callbacks
    glutDisplayFunc(renderScene);
    glutReshapeFunc(changeSize);
    glutKeyboardFunc(processNormalKeys);

    // enter GLUT event processing cycle
    glutMainLoop();
    
    
    return 0;
}

